(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('acorn'), require('magic-string')) :
	typeof define === 'function' && define.amd ? define(['exports', 'acorn', 'magic-string'], factory) :
	(factory((global.buble = global.buble || {}),global.acorn,global.MagicString));
}(this, function (exports,acorn,MagicString) { 'use strict';

	MagicString = 'default' in MagicString ? MagicString['default'] : MagicString;

	var statementsWithBlocks = {
		IfStatement: 'consequent',
		ForStatement: 'body',
		ForInStatement: 'body',
		ForOfStatement: 'body',
		WhileStatement: 'body',
		DoWhileStatement: 'body',
		ArrowFunctionExpression: 'body'
	};

	var Node = function Node ( raw, parent ) {
		Object.defineProperties( this, {
			parent: { value: parent },
			program: { value: parent.program || parent },
			depth: { value: parent.depth + 1 },
			keys: { value: Object.keys( raw ) }
		});

		// special case – body-less if/for/while statements. TODO others?
		var type = statementsWithBlocks[ raw.type ];
		if ( type && raw[ type ].type !== 'BlockStatement' ) {
			var nonBlock = raw[ type ];

			// create a synthetic block statement, otherwise all hell
			// breaks loose when it comes to block scoping
			raw[ type ] = {
				start: nonBlock.start,
				end: nonBlock.end,
				type: 'BlockStatement',
				body: [ nonBlock ],
				synthetic: true
			};
		}

		for ( var i = 0, list = this.keys; i < list.length; i += 1 ) {
			var key = list[i];

				this[ key ] = wrap( raw[ key ], this );
		}

		this.program.magicString.addSourcemapLocation( this.start );
		this.program.magicString.addSourcemapLocation( this.end );
	};

	Node.prototype.ancestor = function ancestor ( level ) {
		var node = this;
		while ( level-- ) {
			node = node.parent;
			if ( !node ) return null;
		}

		return node;
	};

	Node.prototype.contains = function contains ( node ) {
		var this$1 = this;

			while ( node ) {
			if ( node === this$1 ) return true;
			node = node.parent;
		}

		return false;
	};

	Node.prototype.findLexicalBoundary = function findLexicalBoundary () {
		return this.parent.findLexicalBoundary();
	};

	Node.prototype.findNearest = function findNearest ( type ) {
		if ( typeof type === 'string' ) type = new RegExp( ("^" + type + "$") );
		if ( type.test( this.type ) ) return this;
		return this.parent.findNearest( type );
	};

	Node.prototype.findScope = function findScope ( functionScope ) {
		return this.parent.findScope( functionScope );
	};

	Node.prototype.getIndentation = function getIndentation () {
		var lastLine = /\n(.+)$/.exec( this.program.magicString.original.slice( 0, this.start ) );
		return lastLine ? /^[ \t]*/.exec( lastLine[1] )[0] : '';
	};

	Node.prototype.initialise = function initialise ( transforms ) {
		for ( var i = 0, list = this.keys; i < list.length; i += 1 ) {
			var key = list[i];

				var value = this[ key ];

			if ( Array.isArray( value ) ) {
				value.forEach( function ( node ) { return node && node.initialise( transforms ); } );
			} else if ( value && typeof value === 'object' ) {
				value.initialise( transforms );
			}
		}
	};

	Node.prototype.toString = function toString () {
		return this.program.magicString.slice( this.start, this.end );
	};

	Node.prototype.transpile = function transpile ( code, transforms ) {
		for ( var i = 0, list = this.keys; i < list.length; i += 1 ) {
			var key = list[i];

				var value = this[ key ];

			if ( Array.isArray( value ) ) {
				value.forEach( function ( node ) { return node && node.transpile( code, transforms ); } );
			} else if ( value && typeof value === 'object' ) {
				value.transpile( code, transforms );
			}
		}
	};

	var ArrayExpression = (function (Node) {
		function ArrayExpression () {
			Node.apply(this, arguments);
		}

		ArrayExpression.prototype = Object.create( Node && Node.prototype );
		ArrayExpression.prototype.constructor = ArrayExpression;

		ArrayExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.spreadRest ) {
				var lastElement = this.elements[ this.elements.length - 1 ];
				if ( lastElement && lastElement.type === 'SpreadElement' ) {
					var penultimateElement = this.elements[ this.elements.length - 2 ];

					if ( penultimateElement ) {
						code.overwrite( penultimateElement.end, lastElement.start, (" ].concat( ") );
					} else {
						code.insertLeft( this.start + 1, ("].concat(") );
					}

					code.remove( lastElement.start, lastElement.start + 3 );

					var charIndex = lastElement.end;
					while ( code.original[ charIndex ] !== ']' ) charIndex += 1;

					code.overwrite( charIndex, charIndex + 1, ')' );
				}
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return ArrayExpression;
	}(Node));

	var ArrowFunctionExpression = (function (Node) {
		function ArrowFunctionExpression () {
			Node.apply(this, arguments);
		}

		ArrowFunctionExpression.prototype = Object.create( Node && Node.prototype );
		ArrowFunctionExpression.prototype.constructor = ArrowFunctionExpression;

		ArrowFunctionExpression.prototype.initialise = function initialise ( transforms ) {
			this.body.createScope();
			Node.prototype.initialise.call( this, transforms );
		};

		ArrowFunctionExpression.prototype.transpile = function transpile ( code, transforms ) {
			var this$1 = this;

			if ( transforms.arrow ) {
				if ( this.params.length === 0 ) {
					code.overwrite( this.start, this.body.start, 'function () ' );
				} else {
					var parenthesised = false;
					var charIndex = this.start;
					while ( charIndex < this$1.params[0].start ) {
						if ( code.original[ charIndex ] === '(' ) {
							parenthesised = true;
							break;
						}
					}

					var start = 'function ';
					if ( !parenthesised ) start += '( ';

					code.insertRight( this.start, start );

					charIndex = this.params[ this.params.length -1 ].end;
					while ( code.original[ charIndex ] !== '=' ) charIndex += 1;

					// remove the `=> `
					code.overwrite( charIndex, this.body.start, parenthesised ? '' : ') ' );
				}
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return ArrowFunctionExpression;
	}(Node));

	function locate ( source, index ) {
		var lines = source.split( '\n' );
		var len = lines.length;

		var lineStart = 0;
		var i;

		for ( i = 0; i < len; i += 1 ) {
			var line = lines[i];
			var lineEnd =  lineStart + line.length + 1; // +1 for newline

			if ( lineEnd > index ) {
				return { line: i + 1, column: index - lineStart, char: i };
			}

			lineStart = lineEnd;
		}

		throw new Error( 'Could not determine location of character' );
	}

	function pad ( num, len ) {
		var result = String( num );
		return result + repeat( ' ', len - result.length );
	}

	function repeat ( str, times ) {
		var result = '';
		while ( times-- ) result += str;
		return result;
	}

	function getSnippet ( source, loc, length ) {
		if ( length === void 0 ) length = 1;

		var first = Math.max( loc.line - 5, 0 );
		var last = loc.line;

		var numDigits = String( last ).length;

		var lines = source.split( '\n' ).slice( first, last );

		var lastLine = lines[ lines.length - 1 ];
		var offset = lastLine.slice( 0, loc.column ).replace( /\t/g, '  ' ).length;

		var snippet = lines
			.map( function ( line, i ) { return ("" + (pad( i + first + 1, numDigits )) + " : " + (line.replace( /\t/g, '  '))); } )
			.join( '\n' );

		snippet += '\n' + repeat( ' ', numDigits + 3 + offset ) + repeat( '^', length );

		return snippet;
	}

	var CompileError = (function (Error) {
		function CompileError ( node, message ) {
			Error.call(this);

			var source = node.program.magicString.original;
			var loc = locate( source, node.start );

			this.name = 'CompileError';
			this.message = message + " (" + (loc.line) + ":" + (loc.column) + ")";

			this.stack = new Error().stack.replace( new RegExp( (".+new " + (this.name) + ".+\\n"), 'm' ), '' );

			this.loc = loc;
			this.snippet = getSnippet( source, loc, node.end - node.start );
		}

		CompileError.prototype = Object.create( Error && Error.prototype );
		CompileError.prototype.constructor = CompileError;

		return CompileError;
	}(Error));

	var AssignmentExpression = (function (Node) {
		function AssignmentExpression () {
			Node.apply(this, arguments);
		}

		AssignmentExpression.prototype = Object.create( Node && Node.prototype );
		AssignmentExpression.prototype.constructor = AssignmentExpression;

		AssignmentExpression.prototype.initialise = function initialise ( transforms ) {
			if ( this.left.type === 'Identifier' ) {
				var declaration = this.findScope( false ).findDeclaration( this.left.name );
				if ( declaration && declaration.kind === 'const' ) {
					throw new CompileError( this.left, ("" + (this.left.name) + " is read-only") );
				}

				// special case – https://gitlab.com/Rich-Harris/buble/issues/11
				var statement = declaration && declaration.node.ancestor( 3 );
				if ( statement && statement.type === 'ForStatement' && statement.body.contains( this ) ) {
					statement.reassigned[ this.left.name ] = true;
				}
			}

			if ( /Pattern/.test( this.left.type ) ) {
				throw new CompileError( this.left, 'Destructuring assignments are not currently supported. Coming soon!' );
			}

			Node.prototype.initialise.call( this, transforms );
		};

		AssignmentExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( this.operator === '**=' && transforms.exponentiation ) {
				var scope = this.findScope( false );
				var getAlias = function ( name ) {
					var declaration = scope.findDeclaration( name );
					return declaration ? declaration.name : name;
				};

				// first, the easy part – `**=` -> `=`
				var charIndex = this.left.end;
				while ( code.original[ charIndex ] !== '*' ) charIndex += 1;
				code.remove( charIndex, charIndex + 2 );

				// how we do the next part depends on a number of factors – whether
				// this is a top-level statement, and whether we're updating a
				// simple or complex reference
				var base;

				var left = this.left;
				while ( left.type === 'ParenthesizedExpression' ) left = left.expression;

				if ( left.type === 'Identifier' ) {
					base = getAlias( left.name );
				} else if ( left.type === 'MemberExpression' ) {
					var object;
					var needsObjectVar = false;
					var property;
					var needsPropertyVar = false;

					var statement = this.findNearest( /(?:Statement|Declaration)$/ );
					var i0 = statement.getIndentation();

					if ( left.property.type === 'Identifier' ) {
						property = left.computed ? getAlias( left.property.name ) : left.property.name;
					} else {
						property = scope.createIdentifier( 'property' );
						needsPropertyVar = true;
					}

					if ( left.object.type === 'Identifier' ) {
						object = getAlias( left.object.name );
					} else {
						object = scope.createIdentifier( 'object' );
						needsObjectVar = true;
					}

					if ( left.start === statement.start ) {
						if ( needsObjectVar && needsPropertyVar ) {
							code.insertRight( statement.start, ("var " + object + " = ") );
							code.overwrite( left.object.end, left.property.start, (";\n" + i0 + "var " + property + " = ") );
							code.overwrite( left.property.end, left.end, (";\n" + i0 + "" + object + "[" + property + "]") );
						}

						else if ( needsObjectVar ) {
							code.insertRight( statement.start, ("var " + object + " = ") );
							code.insertLeft( left.object.end, (";\n" + i0) );
							code.insertLeft( left.object.end, object );
						}

						else if ( needsPropertyVar ) {
							code.insertRight( left.property.start, ("var " + property + " = ") );
							code.insertLeft( left.property.end, (";\n" + i0) );
							code.move( left.property.start, left.property.end, this.start );

							code.insertLeft( left.object.end, ("[" + property + "]") );
							code.remove( left.object.end, left.property.start );
							code.remove( left.property.end, left.end );
						}
					}

					else {
						var declarators = [];
						if ( needsObjectVar ) declarators.push( object );
						if ( needsPropertyVar ) declarators.push( property );
						code.insertRight( statement.start, ("var " + (declarators.join( ', ' )) + ";\n" + i0) );

						if ( needsObjectVar && needsPropertyVar ) {
							code.insertRight( left.start, ("( " + object + " = ") );
							code.overwrite( left.object.end, left.property.start, (", " + property + " = ") );
							code.overwrite( left.property.end, left.end, (", " + object + "[" + property + "]") );
						}

						else if ( needsObjectVar ) {
							code.insertRight( left.start, ("( " + object + " = ") );
							code.insertLeft( left.object.end, (", " + object) );
						}

						else if ( needsPropertyVar ) {
							code.insertRight( left.property.start, ("( " + property + " = ") );
							code.insertLeft( left.property.end, (", ") );
							code.move( left.property.start, left.property.end, left.start );

							code.overwrite( left.object.end, left.property.start, ("[" + property + "]") );
							code.remove( left.property.end, left.end );
						}

						code.insertLeft( this.end, (" )") );
					}

					base = object + ( left.computed || needsPropertyVar ? ("[" + property + "]") : ("." + property) );
				}

				code.insertRight( this.right.start, ("Math.pow( " + base + ", ") );
				code.insertLeft( this.right.end, (" )") );
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return AssignmentExpression;
	}(Node));

	var BinaryExpression = (function (Node) {
		function BinaryExpression () {
			Node.apply(this, arguments);
		}

		BinaryExpression.prototype = Object.create( Node && Node.prototype );
		BinaryExpression.prototype.constructor = BinaryExpression;

		BinaryExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( this.operator === '**' && transforms.exponentiation ) {
				code.insertRight( this.start, ("Math.pow( ") );
				code.overwrite( this.left.end, this.right.start, (", ") );
				code.insertLeft( this.end, (" )") );
			}
			Node.prototype.transpile.call( this, code, transforms );
		};

		return BinaryExpression;
	}(Node));

	var BreakStatement = (function (Node) {
		function BreakStatement () {
			Node.apply(this, arguments);
		}

		BreakStatement.prototype = Object.create( Node && Node.prototype );
		BreakStatement.prototype.constructor = BreakStatement;

		BreakStatement.prototype.initialise = function initialise ( transforms ) {
			var loop = this.findNearest( /(?:For(?:In)?|While)Statement/ );
			var switchCase = this.findNearest( 'SwitchCase' );

			if ( loop && ( !switchCase || loop.depth > switchCase.depth ) ) {
				loop.canBreak = true;
				this.loop = loop;
			}
		};

		BreakStatement.prototype.transpile = function transpile ( code, transforms ) {
			if ( this.loop && this.loop.shouldRewriteAsFunction ) {
				if ( this.label ) throw new CompileError( this, 'Labels are not currently supported in a loop with locally-scoped variables' );
				code.overwrite( this.start, this.start + 5, ("return 'break'") );
			}
		};

		return BreakStatement;
	}(Node));

	var CallExpression = (function (Node) {
		function CallExpression () {
			Node.apply(this, arguments);
		}

		CallExpression.prototype = Object.create( Node && Node.prototype );
		CallExpression.prototype.constructor = CallExpression;

		CallExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.spreadRest ) {
				var lastArgument = this.arguments[ this.arguments.length - 1 ];
				if ( lastArgument && lastArgument.type === 'SpreadElement' ) {
					var context;

					if ( this.callee.type === 'MemberExpression' ) {
						if ( this.callee.object.type === 'Identifier' ) {
							context = this.callee.object.name;
						} else {
							var statement = this.callee.object;
							var i0 = statement.getIndentation();
							context = this.findScope( true ).createIdentifier( 'ref' );
							code.insertRight( statement.start, ("var " + context + " = ") );
							code.insertLeft( statement.end, (";\n" + i0 + "" + context) );
						}
					} else {
						context = 'void 0';
					}

					code.insertLeft( this.callee.end, '.apply' );

					var penultimateArgument = this.arguments[ this.arguments.length - 2 ];

					if ( penultimateArgument ) {
						code.insertRight( this.arguments[0].start, ("" + context + ", [ ") );
						code.overwrite( penultimateArgument.end, lastArgument.start, (" ].concat( ") );
						code.insertLeft( lastArgument.end, (" )") );
					} else {
						code.insertRight( lastArgument.start, ("" + context + ", ") );
					}

					code.remove( lastArgument.start, lastArgument.start + 3 );
				}
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return CallExpression;
	}(Node));

	function findIndex ( array, fn ) {
		for ( var i = 0; i < array.length; i += 1 ) {
			if ( fn( array[i], i ) ) return i;
		}

		return -1;
	}

	// TODO this code is pretty wild, tidy it up
	var ClassBody = (function (Node) {
		function ClassBody () {
			Node.apply(this, arguments);
		}

		ClassBody.prototype = Object.create( Node && Node.prototype );
		ClassBody.prototype.constructor = ClassBody;

		ClassBody.prototype.transpile = function transpile ( code, transforms, inFunctionExpression, superName ) {
			var this$1 = this;

			if ( transforms.classes ) {
				var name = this.parent.name;

				var indentStr = code.getIndentString();
				var i0 = this.getIndentation() + ( inFunctionExpression ? indentStr : '' );
				var i1 = i0 + indentStr;

				var constructorIndex = findIndex( this.body, function ( node ) { return node.kind === 'constructor'; } );
				var constructor = this.body[ constructorIndex ];

				if ( this.body.length ) {
					code.remove( this.start, this.body[0].start );
					code.remove( this.body[ this.body.length - 1 ].end, this.end );
				} else {
					code.remove( this.start, this.end );
				}

				if ( constructor ) {
					var previousMethod = this.body[ constructorIndex - 1 ];
					var nextMethod = this.body[ constructorIndex + 1 ];

					// ensure constructor is first
					if ( constructorIndex > 0 ) {
						code.remove( previousMethod.end, constructor.start );
						code.move( constructor.start, nextMethod ? nextMethod.start : this.end - 1, this.body[0].start );
					}

					if ( !inFunctionExpression ) code.insertLeft( constructor.end, ';' );
				}

				if ( this.parent.superClass ) {
					var inheritanceBlock = "if ( " + superName + " ) " + name + ".__proto__ = " + superName + ";\n" + i0 + "" + name + ".prototype = Object.create( " + superName + " && " + superName + ".prototype );\n" + i0 + "" + name + ".prototype.constructor = " + name + ";";

					if ( constructor ) {
						code.insertLeft( constructor.end, "\n\n" + i0 + inheritanceBlock );
					} else {
						var fn = "function " + name + " () {" + ( superName ?
							("\n" + i1 + "" + superName + ".apply(this, arguments);\n" + i0 + "}") :
							("}") ) + ( inFunctionExpression ? '' : ';' ) + ( this.body.length ? ("\n\n" + i0) : '' );

						inheritanceBlock = fn + inheritanceBlock;
						code.insertRight( this.start, inheritanceBlock + "\n\n" + i0 );
					}
				} else if ( !constructor ) {
					var fn$1 = "function " + name + " () {}";
					if ( this.parent.type === 'ClassDeclaration' ) fn$1 += ';';
					if ( this.body.length ) fn$1 += "\n\n" + i0;

					code.insertRight( this.start, fn$1 );
				}

				var scope = this.findScope( false );

				var prototypeGettersAndSetters = [];
				var staticGettersAndSetters = [];
				var prototypeAccessors;
				var staticAccessors;

				this.body.forEach( function ( method, i ) {
					if ( method.kind === 'constructor' ) {
						code.overwrite( method.key.start, method.key.end, ("function " + name) );
						return;
					}

					if ( method.static ) code.remove( method.start, method.start + 7 );

					var isAccessor = method.kind !== 'method';
					var lhs;

					if ( isAccessor ) {
						code.remove( method.start, method.key.end );

						if ( method.static ) {
							if ( !~staticGettersAndSetters.indexOf( method.key.name ) ) staticGettersAndSetters.push( method.key.name );
							if ( !staticAccessors ) staticAccessors = scope.createIdentifier( 'staticAccessors' );

							lhs = "" + staticAccessors + "." + (method.key.name) + "." + (method.kind);
						} else {
							if ( !~prototypeGettersAndSetters.indexOf( method.key.name ) ) prototypeGettersAndSetters.push( method.key.name );
							if ( !prototypeAccessors ) prototypeAccessors = scope.createIdentifier( 'prototypeAccessors' );

							lhs = "" + prototypeAccessors + "." + (method.key.name) + "." + (method.kind);
						}
					} else {
						lhs = method.static ?
							("" + name + "." + (method.key.name)) :
							("" + name + ".prototype." + (method.key.name));
					}

					var insertNewlines = ( constructorIndex > 0 && i === constructorIndex + 1 ) ||
					                       ( i === 0 && constructorIndex === this$1.body.length - 1 );

					if ( insertNewlines ) lhs = "\n\n" + i0 + "" + lhs;

					code.insertRight( method.start, "" + lhs + " = function" + ( method.value.generator ? '*' : '' ) + ( isAccessor ? '' : ' ' ) );
					code.insertLeft( method.end, ';' );

					if ( method.value.generator ) code.remove( method.start, method.key.start );

					// prevent function name shadowing an existing declaration
					if ( scope.contains( method.key.name ) ) {
						code.overwrite( method.key.start, method.key.end, scope.createIdentifier( method.key.name ), true );
					}
				});

				if ( prototypeGettersAndSetters.length || staticGettersAndSetters.length ) {
					var intro = [];
					var outro = [];

					if ( prototypeGettersAndSetters.length ) {
						intro.push( ("var " + prototypeAccessors + " = { " + (prototypeGettersAndSetters.map( function ( name ) { return ("" + name + ": {}"); } ).join( ',' )) + " };") );
						outro.push( ("Object.defineProperties( " + name + ".prototype, " + prototypeAccessors + " );") );
					}

					if ( staticGettersAndSetters.length ) {
						intro.push( ("var " + staticAccessors + " = { " + (staticGettersAndSetters.map( function ( name ) { return ("" + name + ": {}"); } ).join( ',' )) + " };") );
						outro.push( ("Object.defineProperties( " + name + ", " + staticAccessors + " );") );
					}

					if ( constructor ) {
						code.insertLeft( constructor.end, ("\n\n" + i0 + "" + (intro.join( ("\n" + i0) ))) );
					} else {
						code.insertRight( this.start, ("" + (intro.join( ("\n" + i0) )) + "\n\n" + i0) );
					}

					code.insertLeft( this.end, ("\n\n" + i0 + "" + (outro.join( ("\n" + i0) ))) );
				}
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return ClassBody;
	}(Node));

	// TODO this function is slightly flawed – it works on the original string,
	// not its current edited state.
	// That's not a problem for the way that it's currently used, but it could
	// be in future...
	function deindent ( node, code ) {
		var start = node.start;
		var end = node.end;

		var indentStr = code.getIndentString();
		var pattern = new RegExp( indentStr + '\\S', 'g' );

		if ( code.original.slice( start - indentStr.length, start ) === indentStr ) {
			code.remove( start - indentStr.length, start );
		}

		var slice = code.original.slice( start, end );
		var match;
		while ( match = pattern.exec( slice ) ) {
			if ( !node.program.indentExclusions[ match.index ] ) code.remove( start + match.index, start + match.index + indentStr.length );
		}
	}

	var ClassDeclaration = (function (Node) {
		function ClassDeclaration () {
			Node.apply(this, arguments);
		}

		ClassDeclaration.prototype = Object.create( Node && Node.prototype );
		ClassDeclaration.prototype.constructor = ClassDeclaration;

		ClassDeclaration.prototype.initialise = function initialise ( transforms ) {
			this.name = this.id.name;
			this.findScope( true ).addDeclaration( this.id, 'class' );

			Node.prototype.initialise.call( this, transforms );
		};

		ClassDeclaration.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.classes ) {
				if ( !this.superClass ) deindent( this.body, code );

				var superName = this.superClass && ( this.superClass.name || 'superclass' );

				var i0 = this.getIndentation();
				var i1 = i0 + code.getIndentString();

				// if this is an export default statement, we have to move the export to
				// after the declaration, because `export default var Foo = ...` is illegal
				var syntheticDefaultExport = this.parent.type === 'ExportDefaultDeclaration' ?
					("\n\n" + i0 + "export default " + (this.id.name) + ";") :
					'';

				if ( syntheticDefaultExport ) code.remove( this.parent.start, this.start );

				code.overwrite( this.start, this.id.start, 'var ' );

				if ( this.superClass ) {
					code.overwrite( this.id.end, this.superClass.start, ' = ' );
					code.overwrite( this.superClass.end, this.body.start, ("(function (" + superName + ") {\n" + i1) );
				} else {
					code.overwrite( this.id.end, this.body.start, ' = ' );
				}

				this.body.transpile( code, transforms, !!this.superClass, superName );

				if ( this.superClass ) {
					code.insertLeft( this.end, ("\n\n" + i1 + "return " + (this.name) + ";\n" + i0 + "}(") );
					code.move( this.superClass.start, this.superClass.end, this.end );
					code.insertRight( this.end, ("));" + syntheticDefaultExport) );
				} else if ( syntheticDefaultExport ) {
					code.insertRight( this.end, syntheticDefaultExport );
				}
			}

			else {
				this.body.transpile( code, transforms, false, null );
			}
		};

		return ClassDeclaration;
	}(Node));

	var ClassExpression = (function (Node) {
		function ClassExpression () {
			Node.apply(this, arguments);
		}

		ClassExpression.prototype = Object.create( Node && Node.prototype );
		ClassExpression.prototype.constructor = ClassExpression;

		ClassExpression.prototype.initialise = function initialise ( transforms ) {
			this.name = this.id ? this.id.name :
			            this.parent.type === 'VariableDeclarator' ? this.parent.id.name :
			            this.parent.type === 'AssignmentExpression' ? this.parent.left.name :
			            this.findScope( true ).createIdentifier( 'anonymous' );

			Node.prototype.initialise.call( this, transforms );
		};

		ClassExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.classes ) {
				var superName = this.superClass && ( this.superClass.name || 'superclass' );

				var i0 = this.getIndentation();
				var i1 = i0 + code.getIndentString();

				if ( this.superClass ) {
					code.remove( this.start, this.superClass.start );
					code.remove( this.superClass.end, this.body.start );
					code.insertLeft( this.start, ("(function (" + superName + ") {\n" + i1) );
				} else {
					code.overwrite( this.start, this.body.start, ("(function () {\n" + i1) );
				}

				this.body.transpile( code, transforms, true, superName );

				var outro = "\n\n" + i1 + "return " + (this.name) + ";\n" + i0 + "}(";

				if ( this.superClass ) {
					code.insertLeft( this.end, outro );
					code.move( this.superClass.start, this.superClass.end, this.end );
					code.insertRight( this.end, '))' );
				} else {
					code.insertLeft( this.end, ("\n\n" + i1 + "return " + (this.name) + ";\n" + i0 + "}())") );
				}
			}

			else {
				this.body.transpile( code, transforms, false );
			}
		};

		return ClassExpression;
	}(Node));

	var ContinueStatement = (function (Node) {
		function ContinueStatement () {
			Node.apply(this, arguments);
		}

		ContinueStatement.prototype = Object.create( Node && Node.prototype );
		ContinueStatement.prototype.constructor = ContinueStatement;

		ContinueStatement.prototype.transpile = function transpile ( code, transforms ) {
			var loop = this.findNearest( /(?:For(?:In)?|While)Statement/ );
			if ( loop.shouldRewriteAsFunction ) {
				if ( this.label ) throw new CompileError( this, 'Labels are not currently supported in a loop with locally-scoped variables' );
				code.overwrite( this.start, this.start + 8, 'return' );
			}
		};

		return ContinueStatement;
	}(Node));

	var ExportDefaultDeclaration = (function (Node) {
		function ExportDefaultDeclaration () {
			Node.apply(this, arguments);
		}

		ExportDefaultDeclaration.prototype = Object.create( Node && Node.prototype );
		ExportDefaultDeclaration.prototype.constructor = ExportDefaultDeclaration;

		ExportDefaultDeclaration.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.moduleExport ) throw new CompileError( this, 'export is not supported' );
			Node.prototype.initialise.call( this, transforms );
		};

		return ExportDefaultDeclaration;
	}(Node));

	var ExportNamedDeclaration = (function (Node) {
		function ExportNamedDeclaration () {
			Node.apply(this, arguments);
		}

		ExportNamedDeclaration.prototype = Object.create( Node && Node.prototype );
		ExportNamedDeclaration.prototype.constructor = ExportNamedDeclaration;

		ExportNamedDeclaration.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.moduleExport ) throw new CompileError( this, 'export is not supported' );
			Node.prototype.initialise.call( this, transforms );
		};

		return ExportNamedDeclaration;
	}(Node));

	function extractNames ( node ) {
		var names = [];
		extractors[ node.type ]( names, node );
		return names;
	}

	var extractors = {
		Identifier: function Identifier ( names, param ) {
			names.push( param.name );
		},

		ObjectPattern: function ObjectPattern ( names, param ) {
			for ( var i = 0, list = param.properties; i < list.length; i += 1 ) {
				var prop = list[i];

				extractors[ prop.value.type ]( names, prop.value );
			}
		},

		ArrayPattern: function ArrayPattern ( names, param ) {
			for ( var i = 0, list = param.elements; i < list.length; i += 1 )  {
				var element = list[i];

				if ( element ) extractors[ element.type ]( names, element );
			}
		},

		RestElement: function RestElement ( names, param ) {
			extractors[ param.argument.type ]( names, param.argument );
		},

		AssignmentPattern: function AssignmentPattern ( names, param ) {
			extractors[ param.left.type ]( names, param.left );
		}
	};

	var LoopStatement = (function (Node) {
		function LoopStatement () {
			Node.apply(this, arguments);
		}

		LoopStatement.prototype = Object.create( Node && Node.prototype );
		LoopStatement.prototype.constructor = LoopStatement;

		LoopStatement.prototype.findScope = function findScope ( functionScope ) {
			return functionScope || !this.createdScope ? this.parent.findScope( functionScope ) : this.body.scope;
		};

		LoopStatement.prototype.initialise = function initialise ( transforms ) {
			var this$1 = this;

			this.body.createScope();
			this.createdScope = true;

			// this is populated as and when reassignments occur
			this.reassigned = Object.create( null );
			this.aliases = Object.create( null );

			Node.prototype.initialise.call( this, transforms );

			if ( transforms.letConst ) {
				// see if any block-scoped declarations are referenced
				// inside function expressions
				var names = Object.keys( this.body.scope.declarations );

				var i = names.length;
				while ( i-- ) {
					var name = names[i];
					var declaration = this$1.body.scope.declarations[ name ];

					var j = declaration.instances.length;
					while ( j-- ) {
						var instance = declaration.instances[j];
						var nearestFunctionExpression = instance.findNearest( /Function/ );

						if ( nearestFunctionExpression && nearestFunctionExpression.depth > this$1.depth ) {
							this$1.shouldRewriteAsFunction = true;
							break;
						}
					}

					if ( this$1.shouldRewriteAsFunction ) break;
				}
			}
		};

		LoopStatement.prototype.transpile = function transpile ( code, transforms ) {
			if ( this.shouldRewriteAsFunction ) {
				var i0 = this.getIndentation();
				var i1 = i0 + code.getIndentString();

				var argString = this.args ? (" " + (this.args.join( ', ' )) + " ") : '';
				var paramString = this.params ? (" " + (this.params.join( ', ' )) + " ") : '';

				var functionScope = this.findScope( true );
				var loop = functionScope.createIdentifier( 'loop' );

				var before = "var " + loop + " = function (" + paramString + ") " + ( this.body.synthetic ? ("{\n" + i0 + "" + (code.getIndentString())) : '' );
				var after = ( this.body.synthetic ? ("\n" + i0 + "}") : '' ) + ";\n\n" + i0;

				code.insertRight( this.body.start, before );
				code.insertLeft( this.body.end, after );
				code.move( this.start, this.body.start, this.body.end );

				if ( this.canBreak || this.canReturn ) {
					var returned = functionScope.createIdentifier( 'returned' );

					var insert = "{\n" + i1 + "var " + returned + " = " + loop + "(" + argString + ");\n";
					if ( this.canBreak ) insert += "\n" + i1 + "if ( " + returned + " === 'break' ) break;";
					if ( this.canReturn ) insert += "\n" + i1 + "if ( " + returned + " ) return returned.v;";
					insert += "\n" + i0 + "}";

					code.insertRight( this.body.end, insert );
				} else {
					var callExpression = "" + loop + "(" + argString + ");";

					if ( this.type === 'DoWhileStatement' ) {
						code.overwrite( this.start, this.body.start, ("do {\n" + i1 + "" + callExpression + "\n" + i0 + "}") );
					} else {
						code.insertRight( this.body.end, callExpression );
					}
				}
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return LoopStatement;
	}(Node));

	var ForStatement = (function (LoopStatement) {
		function ForStatement () {
			LoopStatement.apply(this, arguments);
		}

		ForStatement.prototype = Object.create( LoopStatement && LoopStatement.prototype );
		ForStatement.prototype.constructor = ForStatement;

		ForStatement.prototype.findScope = function findScope ( functionScope ) {
			return functionScope || !this.createdScope ? this.parent.findScope( functionScope ) : this.body.scope;
		};

		ForStatement.prototype.transpile = function transpile ( code, transforms ) {
			var this$1 = this;

			var i1 = this.getIndentation() + code.getIndentString();

			if ( this.shouldRewriteAsFunction ) {
				// which variables are declared in the init statement?
				var names = this.init.type === 'VariableDeclaration' ?
					[].concat.apply( [], this.init.declarations.map( function ( declarator ) { return extractNames( declarator.id ); } ) ) :
					[];

				var aliases = this.aliases;

				this.args = names.map( function ( name ) { return name in this$1.aliases ? this$1.aliases[ name ].outer : name; } );
				this.params = names.map( function ( name ) { return name in this$1.aliases ? this$1.aliases[ name ].inner : name; } );

				var updates = Object.keys( this.reassigned )
					.map( function ( name ) { return ("" + (aliases[ name ].outer) + " = " + (aliases[ name ].inner) + ";"); } );

				if ( updates.length ) {
					if ( this.body.synthetic ) {
						code.insertLeft( this.body.body[0].end, ("; " + (updates.join((" ")))) );
					} else {
						var lastStatement = this.body.body[ this.body.body.length - 1 ];
						code.insertLeft( lastStatement.end, ("\n\n" + i1 + "" + (updates.join(("\n" + i1)))) );
					}
				}
			}

			LoopStatement.prototype.transpile.call( this, code, transforms );
		};

		return ForStatement;
	}(LoopStatement));

	var ForInStatement = (function (LoopStatement) {
		function ForInStatement () {
			LoopStatement.apply(this, arguments);
		}

		ForInStatement.prototype = Object.create( LoopStatement && LoopStatement.prototype );
		ForInStatement.prototype.constructor = ForInStatement;

		ForInStatement.prototype.findScope = function findScope ( functionScope ) {
			return functionScope || !this.createdScope ? this.parent.findScope( functionScope ) : this.body.scope;
		};

		ForInStatement.prototype.transpile = function transpile ( code, transforms ) {
			var this$1 = this;

			if ( this.shouldRewriteAsFunction ) {
				// which variables are declared in the init statement?
				var names = this.left.type === 'VariableDeclaration' ?
					[].concat.apply( [], this.left.declarations.map( function ( declarator ) { return extractNames( declarator.id ); } ) ) :
					[];

				this.args = names.map( function ( name ) { return name in this$1.aliases ? this$1.aliases[ name ].outer : name; } );
				this.params = names.map( function ( name ) { return name in this$1.aliases ? this$1.aliases[ name ].inner : name; } );
			}

			LoopStatement.prototype.transpile.call( this, code, transforms );
		};

		return ForInStatement;
	}(LoopStatement));

	var ForOfStatement = (function (LoopStatement) {
		function ForOfStatement () {
			LoopStatement.apply(this, arguments);
		}

		ForOfStatement.prototype = Object.create( LoopStatement && LoopStatement.prototype );
		ForOfStatement.prototype.constructor = ForOfStatement;

		ForOfStatement.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.forOf && !transforms.dangerousForOf ) throw new CompileError( this, 'for...of statements are not supported. Use `transforms: { forOf: false }` to skip transformation and disable this error, or `transforms: { dangerousForOf: true }` if you know what you\'re doing' );
			LoopStatement.prototype.initialise.call( this, transforms );
		};

		ForOfStatement.prototype.transpile = function transpile ( code, transforms ) {
			if ( !transforms.dangerousForOf ) {
				LoopStatement.prototype.transpile.call( this, code, transforms );
				return;
			}

			var scope = this.findScope( true );
			var i0 = this.getIndentation();
			var i1 = i0 + code.getIndentString();

			var key = scope.createIdentifier( 'i' );
			var list = scope.createIdentifier( 'list' );

			if ( this.body.synthetic ) {
				code.insertRight( this.left.start, ("{\n" + i1) );
				code.insertLeft( this.body.body[0].end, ("\n" + i0 + "}") );
			}

			var bodyStart = this.body.body[0].start;

			code.remove( this.left.end, this.right.start );
			code.move( this.left.start, this.left.end, bodyStart );
			code.insertLeft( this.left.end, (" = " + list + "[" + key + "];\n\n" + i1) );

			code.insertRight( this.right.start, ("var " + key + " = 0, " + list + " = ") );
			code.insertLeft( this.right.end, ("; " + key + " < " + list + ".length; " + key + " += 1") );

			LoopStatement.prototype.transpile.call( this, code, transforms );
		};

		return ForOfStatement;
	}(LoopStatement));

	var FunctionDeclaration = (function (Node) {
		function FunctionDeclaration () {
			Node.apply(this, arguments);
		}

		FunctionDeclaration.prototype = Object.create( Node && Node.prototype );
		FunctionDeclaration.prototype.constructor = FunctionDeclaration;

		FunctionDeclaration.prototype.initialise = function initialise ( transforms ) {
			if ( this.generator && transforms.generator ) {
				throw new CompileError( this, 'Generators are not supported' );
			}

			this.body.createScope();

			this.findScope( true ).addDeclaration( this.id, 'function' );
			Node.prototype.initialise.call( this, transforms );
		};

		return FunctionDeclaration;
	}(Node));

	var FunctionExpression = (function (Node) {
		function FunctionExpression () {
			Node.apply(this, arguments);
		}

		FunctionExpression.prototype = Object.create( Node && Node.prototype );
		FunctionExpression.prototype.constructor = FunctionExpression;

		FunctionExpression.prototype.initialise = function initialise ( transforms ) {
			if ( this.generator && transforms.generator ) {
				throw new CompileError( this, 'Generators are not supported' );
			}

			this.body.createScope();

			if ( this.id ) {
				// function expression IDs belong to the child scope...
				this.body.scope.addDeclaration( this.id, 'function' );
			}

			Node.prototype.initialise.call( this, transforms );
		};

		return FunctionExpression;
	}(Node));

	function isReference ( node, parent ) {
		if ( node.type === 'MemberExpression' ) {
			return !node.computed && isReference( node.object, node );
		}

		if ( node.type === 'Identifier' ) {
			// the only time we could have an identifier node without a parent is
			// if it's the entire body of a function without a block statement –
			// i.e. an arrow function expression like `a => a`
			if ( !parent ) return true;

			if ( /(Function|Class)Expression/.test( parent.type ) ) return false;

			// TODO is this right?
			if ( parent.type === 'MemberExpression' || parent.type === 'MethodDefinition' ) {
				return parent.computed || node === parent.object;
			}

			// disregard the `bar` in `{ bar: foo }`, but keep it in `{ [bar]: foo }`
			if ( parent.type === 'Property' ) return parent.computed || node === parent.value;

			// disregard the `bar` in `class Foo { bar () {...} }`
			if ( parent.type === 'MethodDefinition' ) return false;

			// disregard the `bar` in `export { foo as bar }`
			if ( parent.type === 'ExportSpecifier' && node !== parent.local ) return false;

			return true;
		}
	}

	var Identifier = (function (Node) {
		function Identifier () {
			Node.apply(this, arguments);
		}

		Identifier.prototype = Object.create( Node && Node.prototype );
		Identifier.prototype.constructor = Identifier;

		Identifier.prototype.findScope = function findScope ( functionScope ) {
			if ( this.parent.params && ~this.parent.params.indexOf( this ) ) {
				return this.parent.body.scope;
			}

			if ( this.parent.type === 'FunctionExpression' && this === this.parent.id ) {
				return this.parent.body.scope;
			}

			return this.parent.findScope( functionScope	);
		};

		Identifier.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.arrow && isReference( this, this.parent ) ) {
				if ( this.name === 'arguments' && !this.findScope( false ).contains( this.name ) ) {
					var lexicalBoundary = this.findLexicalBoundary();
					var arrowFunction = this.findNearest( 'ArrowFunctionExpression' );
					var loop = this.findNearest( /(?:For|While)Statement/ );

					if ( arrowFunction && arrowFunction.depth > lexicalBoundary.depth ) {
						this.alias = lexicalBoundary.getArgumentsAlias();
					}

					if ( loop && loop.depth > lexicalBoundary.depth ) {
						this.alias = lexicalBoundary.getArgumentsAlias();
					}
				}

				this.findScope( false ).addReference( this );
			}
		};

		Identifier.prototype.transpile = function transpile ( code, transforms ) {
			if ( this.alias ) {
				code.overwrite( this.start, this.end, this.alias, true );
			}
		};

		return Identifier;
	}(Node));

	var ImportDeclaration = (function (Node) {
		function ImportDeclaration () {
			Node.apply(this, arguments);
		}

		ImportDeclaration.prototype = Object.create( Node && Node.prototype );
		ImportDeclaration.prototype.constructor = ImportDeclaration;

		ImportDeclaration.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.moduleImport ) throw new CompileError( this, 'import is not supported' );
			Node.prototype.initialise.call( this, transforms );
		};

		return ImportDeclaration;
	}(Node));

	var ImportDefaultSpecifier = (function (Node) {
		function ImportDefaultSpecifier () {
			Node.apply(this, arguments);
		}

		ImportDefaultSpecifier.prototype = Object.create( Node && Node.prototype );
		ImportDefaultSpecifier.prototype.constructor = ImportDefaultSpecifier;

		ImportDefaultSpecifier.prototype.initialise = function initialise ( transforms ) {
			this.findScope( true ).addDeclaration( this.local, 'import' );
			Node.prototype.initialise.call( this, transforms );
		};

		return ImportDefaultSpecifier;
	}(Node));

	var ImportSpecifier = (function (Node) {
		function ImportSpecifier () {
			Node.apply(this, arguments);
		}

		ImportSpecifier.prototype = Object.create( Node && Node.prototype );
		ImportSpecifier.prototype.constructor = ImportSpecifier;

		ImportSpecifier.prototype.initialise = function initialise ( transforms ) {
			this.findScope( true ).addDeclaration( this.local, 'import' );
			Node.prototype.initialise.call( this, transforms );
		};

		return ImportSpecifier;
	}(Node));

	var Literal = (function (Node) {
		function Literal () {
			Node.apply(this, arguments);
		}

		Literal.prototype = Object.create( Node && Node.prototype );
		Literal.prototype.constructor = Literal;

		Literal.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.numericLiteral ) {
				var leading = this.raw.slice( 0, 2 );
				if ( leading === '0b' || leading === '0o' ) {
					code.overwrite( this.start, this.end, String( this.value ), true );
				}
			}

			if ( this.regex ) {
				if ( transforms.unicodeRegExp && /u/.test( this.regex.flags ) ) throw new CompileError( this, 'Regular expression unicode flag is not supported' );
				if ( transforms.stickyRegExp && /y/.test( this.regex.flags ) ) throw new CompileError( this, 'Regular expression sticky flag is not supported' );
			}
		};

		return Literal;
	}(Node));

	var reserved = Object.create( null );
	'do if in for let new try var case else enum eval null this true void with await break catch class const false super throw while yield delete export import public return static switch typeof default extends finally package private continue debugger function arguments interface protected implements instanceof'.split( ' ' )
		.forEach( function ( word ) { return reserved[ word ] = true; } );

	var MemberExpression = (function (Node) {
		function MemberExpression () {
			Node.apply(this, arguments);
		}

		MemberExpression.prototype = Object.create( Node && Node.prototype );
		MemberExpression.prototype.constructor = MemberExpression;

		MemberExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.reservedProperties && reserved[ this.property.name ] ) {
				code.overwrite( this.object.end, this.property.start, ("['") );
				code.insertLeft( this.property.end, ("']") );
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return MemberExpression;
	}(Node));

	var Property = (function (Node) {
		function Property () {
			Node.apply(this, arguments);
		}

		Property.prototype = Object.create( Node && Node.prototype );
		Property.prototype.constructor = Property;

		Property.prototype.initialise = function initialise ( transforms ) {
			if ( this.computed && transforms.computedProperty ) {
				throw new CompileError( this.key, 'Computed properties are not supported' );
			}

			Node.prototype.initialise.call( this, transforms );
		};

		Property.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.conciseMethodProperty && this.parent.type !== 'ObjectPattern' ) {
				if ( this.shorthand ) {
					code.insertRight( this.start, ("" + (this.key.name) + ": ") );
				} else if ( this.method ) {
					var name = this.findScope( true ).createIdentifier( this.key.name );
					if ( this.value.generator ) code.remove( this.start, this.key.start );
					code.insertLeft( this.key.end, (": function" + (this.value.generator ? '*' : '') + " " + name) );
				}
			}

			if ( transforms.reservedProperties && reserved[ this.key.name ] ) {
				code.insertRight( this.key.start, ("'") );
				code.insertLeft( this.key.end, ("'") );
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return Property;
	}(Node));

	var ReturnStatement = (function (Node) {
		function ReturnStatement () {
			Node.apply(this, arguments);
		}

		ReturnStatement.prototype = Object.create( Node && Node.prototype );
		ReturnStatement.prototype.constructor = ReturnStatement;

		ReturnStatement.prototype.initialise = function initialise ( transforms ) {
			this.loop = this.findNearest( /(?:For(?:In)?|While)Statement/ );
			this.nearestFunction = this.findNearest( /Function/ );

			if ( this.loop && ( !this.nearestFunction || this.loop.depth > this.nearestFunction.depth ) ) {
				this.loop.canReturn = true;
				this.shouldWrap = true;
			}

			if ( this.argument ) this.argument.initialise( transforms );
		};

		ReturnStatement.prototype.transpile = function transpile ( code, transforms ) {
			if ( this.argument ) {
				var shouldWrap = this.shouldWrap && this.loop && this.loop.shouldRewriteAsFunction;
				if ( shouldWrap ) code.insertRight( this.argument.start, ("{ v: ") );

				if ( this.argument ) this.argument.transpile( code, transforms );

				if ( shouldWrap ) code.insertLeft( this.argument.end, (" }") );
			}
		};

		return ReturnStatement;
	}(Node));

	var Super = (function (Node) {
		function Super () {
			Node.apply(this, arguments);
		}

		Super.prototype = Object.create( Node && Node.prototype );
		Super.prototype.constructor = Super;

		Super.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.classes ) {
				this.method = this.findNearest( 'MethodDefinition' );
				if ( !this.method ) throw new CompileError( this, 'use of super outside class method' );

				var parentClass = this.findNearest( 'ClassBody' ).parent;
				this.superClassName = parentClass.superClass && (parentClass.superClass.name || 'superclass');

				if ( !this.superClassName ) throw new CompileError( this, 'super used in base class' );

				this.isCalled = this.parent.type === 'CallExpression' && this === this.parent.callee;

				if ( this.method.kind !== 'constructor' && this.isCalled ) {
					throw new CompileError( this, 'super() not allowed outside class constructor' );
				}

				this.isMember = this.parent.type === 'MemberExpression';

				if ( !this.isCalled && !this.isMember ) {
					throw new CompileError( this, 'Unexpected use of `super` (expected `super(...)` or `super.*`)' );
				}
			}
		};

		Super.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.classes ) {
				var expression = ( this.isCalled || this.method.static ) ?
					this.superClassName :
					("" + (this.superClassName) + ".prototype");

				code.overwrite( this.start, this.end, expression, true );

				var callExpression = this.isCalled ? this.parent : this.parent.parent;

				if ( callExpression && callExpression.type === 'CallExpression' ) {
					code.insertLeft( callExpression.callee.end, '.call' );

					if ( callExpression.arguments.length ) {
						code.insertLeft( callExpression.arguments[0].start, ("this, ") );
					} else {
						code.insertLeft( callExpression.end - 1, ("this") );
					}
				}
			}
		};

		return Super;
	}(Node));

	var TaggedTemplateExpression = (function (Node) {
		function TaggedTemplateExpression () {
			Node.apply(this, arguments);
		}

		TaggedTemplateExpression.prototype = Object.create( Node && Node.prototype );
		TaggedTemplateExpression.prototype.constructor = TaggedTemplateExpression;

		TaggedTemplateExpression.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.templateString && !transforms.dangerousTaggedTemplateString ) {
				throw new CompileError( this, 'Tagged template strings are not supported. Use `transforms: { templateString: false }` to skip transformation and disable this error, or `transforms: { dangerousTaggedTemplateString: true }` if you know what you\'re doing' );
			}

			Node.prototype.initialise.call( this, transforms );
		};

		TaggedTemplateExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.templateString && transforms.dangerousTaggedTemplateString ) {
				var ordered = this.quasi.expressions.concat( this.quasi.quasis ).sort( function ( a, b ) { return a.start - b.start; } );

				// insert strings at start
				var templateStrings = this.quasi.quasis.map( function ( quasi ) { return JSON.stringify( quasi.value.cooked ); } );
				code.overwrite( this.tag.end, ordered[0].start, ("([" + (templateStrings.join(', ')) + "]") );

				var lastIndex = ordered[0].start;
				ordered.forEach( function ( node ) {
					if ( node.type === 'TemplateElement' ) {
						code.remove( lastIndex, node.end );
					} else {
						code.overwrite( lastIndex, node.start, ', ' );
					}

					lastIndex = node.end;
				});

				code.overwrite( lastIndex, this.end, ')' );
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return TaggedTemplateExpression;
	}(Node));

	var TemplateElement = (function (Node) {
		function TemplateElement () {
			Node.apply(this, arguments);
		}

		TemplateElement.prototype = Object.create( Node && Node.prototype );
		TemplateElement.prototype.constructor = TemplateElement;

		TemplateElement.prototype.initialise = function initialise ( transforms ) {
			this.program.templateElements.push( this );
		};

		return TemplateElement;
	}(Node));

	var TemplateLiteral = (function (Node) {
		function TemplateLiteral () {
			Node.apply(this, arguments);
		}

		TemplateLiteral.prototype = Object.create( Node && Node.prototype );
		TemplateLiteral.prototype.constructor = TemplateLiteral;

		TemplateLiteral.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.templateString && this.parent.type !== 'TaggedTemplateExpression' ) {
				var ordered = this.expressions.concat( this.quasis )
					.sort( function ( a, b ) { return a.start - b.start || a.end - b.end; } )
					.filter( function ( node, i ) {
						// include all expressions
						if ( node.type !== 'TemplateElement' ) return true;

						// include all non-empty strings
						if ( node.value.raw ) return true;

						// exclude all empty strings not at the head
						return !i;
					});

				// special case – we may be able to skip the first element,
				// if it's the empty string, but only if the second and
				// third elements aren't both expressions (since they maybe
				// be numeric, and `1 + 2 + '3' === '33'`)
				if ( ordered.length >= 3 ) {
					var first = ordered[0], third = ordered[2];
					if ( first.type === 'TemplateElement' && first.value.raw === '' && third.type === 'TemplateElement' ) {
						ordered.shift();
					}
				}

				var parenthesise = ( this.quasis.length !== 1 || this.expressions.length !== 0 ) &&
				                     this.parent.type !== 'AssignmentExpression' &&
				                     this.parent.type !== 'VariableDeclarator' &&
				                     ( this.parent.type !== 'BinaryExpression' || this.parent.operator !== '+' );

				if ( parenthesise ) code.insertRight( this.start, '(' );

				var lastIndex = this.start;

				ordered.forEach( function ( node, i ) {
					if ( node.type === 'TemplateElement' ) {
						var replacement = '';
						if ( i ) replacement += ' + ';
						replacement += JSON.stringify( node.value.cooked );

						code.overwrite( lastIndex, node.end, replacement );
					} else {
						var parenthesise = node.type !== 'Identifier'; // TODO other cases where it's safe

						var replacement$1 = '';
						if ( i ) replacement$1 += ' + ';
						if ( parenthesise ) replacement$1 += '(';

						code.overwrite( lastIndex, node.start, replacement$1 );

						if ( parenthesise ) code.insertLeft( node.end, ')' );
					}

					lastIndex = node.end;
				});

				var close = '';
				if ( parenthesise ) close += ')';

				code.overwrite( lastIndex, this.end, close );
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return TemplateLiteral;
	}(Node));

	var ThisExpression = (function (Node) {
		function ThisExpression () {
			Node.apply(this, arguments);
		}

		ThisExpression.prototype = Object.create( Node && Node.prototype );
		ThisExpression.prototype.constructor = ThisExpression;

		ThisExpression.prototype.initialise = function initialise ( transforms ) {
			if ( transforms.arrow ) {
				var lexicalBoundary = this.findLexicalBoundary();
				var arrowFunction = this.findNearest( 'ArrowFunctionExpression' );
				var loop = this.findNearest( /(?:For|While)Statement/ );

				if ( arrowFunction && arrowFunction.depth > lexicalBoundary.depth ) {
					this.alias = lexicalBoundary.getThisAlias();
				}

				if ( loop && loop.depth > lexicalBoundary.depth ) {
					this.alias = lexicalBoundary.getThisAlias();
				}
			}
		};

		ThisExpression.prototype.transpile = function transpile ( code, transforms ) {
			if ( this.alias ) {
				code.overwrite( this.start, this.end, this.alias, true );
			}
		};

		return ThisExpression;
	}(Node));

	var UpdateExpression = (function (Node) {
		function UpdateExpression () {
			Node.apply(this, arguments);
		}

		UpdateExpression.prototype = Object.create( Node && Node.prototype );
		UpdateExpression.prototype.constructor = UpdateExpression;

		UpdateExpression.prototype.initialise = function initialise ( transforms ) {
			if ( this.argument.type === 'Identifier' ) {
				var declaration = this.findScope( false ).findDeclaration( this.argument.name );
				if ( declaration && declaration.kind === 'const' ) {
					throw new CompileError( this, ("" + (this.argument.name) + " is read-only") );
				}
			}

			Node.prototype.initialise.call( this, transforms );
		};

		return UpdateExpression;
	}(Node));

	var VariableDeclaration = (function (Node) {
		function VariableDeclaration () {
			Node.apply(this, arguments);
		}

		VariableDeclaration.prototype = Object.create( Node && Node.prototype );
		VariableDeclaration.prototype.constructor = VariableDeclaration;

		VariableDeclaration.prototype.initialise = function initialise ( transforms ) {
			this.scope = this.findScope( this.kind === 'var' );
			this.declarations.forEach( function ( declarator ) { return declarator.initialise( transforms ); } );
		};

		VariableDeclaration.prototype.transpile = function transpile ( code, transforms ) {
			if ( transforms.letConst && this.kind !== 'var' ) {
				code.overwrite( this.start, this.start + this.kind.length, 'var', true );
			}

			this.declarations.forEach( function ( declarator ) { return declarator.transpile( code, transforms ); } );
		};

		return VariableDeclaration;
	}(Node));

	var VariableDeclarator = (function (Node) {
		function VariableDeclarator () {
			Node.apply(this, arguments);
		}

		VariableDeclarator.prototype = Object.create( Node && Node.prototype );
		VariableDeclarator.prototype.constructor = VariableDeclarator;

		VariableDeclarator.prototype.initialise = function initialise ( transforms ) {
			var this$1 = this;

			this.isObjectPattern = this.id.type === 'ObjectPattern';

			// disallow compound destructuring, for now at least
			if ( /Pattern/.test( this.id.type ) ) {
				this.id[ this.isObjectPattern ? 'properties' : 'elements' ].forEach( function ( node ) {
					if ( node && /Pattern/.test( this$1.isObjectPattern ? node.value.type : node.type ) ) {
						throw new CompileError( node.value, 'Compound destructuring is not supported' );
					}
				});
			}

			var kind = this.parent.kind;
			if ( kind === 'let' && this.parent.parent.type === 'ForStatement' ) {
				kind = 'for.let'; // special case...
			}

			this.parent.scope.addDeclaration( this.id, kind );
			Node.prototype.initialise.call( this, transforms );
		};

		VariableDeclarator.prototype.transpile = function transpile ( code, transforms ) {
			var this$1 = this;

			if ( transforms.destructuring && this.id.type !== 'Identifier' ) {
				var simple = this.init.type === 'Identifier' && !this.init.rewritten;
				var props = this.isObjectPattern ? this.id.properties : this.id.elements;
				var i0 = this.getIndentation();

				var name = simple ? this.init.name : this.findScope( true ).createIdentifier( 'ref' );

				var c = this.start;
				var first = simple;

				if ( simple ) {
					code.remove( this.id.end, this.end );
				} else {
					code.insertRight( this.id.end, ("" + name) );
					code.move( this.id.end, this.end, c );
				}

				props.forEach( function ( property, i ) {
					if ( property ) {
						var id = this$1.isObjectPattern ? property.value : property;
						var rhs = this$1.isObjectPattern ? ("" + name + "." + (property.key.name)) : ("" + name + "[" + i + "]");

						var start = first ? '' : 'var ';
						if ( !first ) start = ";\n" + i0 + "" + start;

						code.insertRight( id.start, start );
						code.move( id.start, id.end, this$1.start );
						code.insertLeft( id.end, (" = " + rhs) );

						code.remove( c, id.start );
						c = property.end;

						first = false;
					}
				});

				code.remove( c, this.id.end );
			}

			Node.prototype.transpile.call( this, code, transforms );
		};

		return VariableDeclarator;
	}(Node));

	var types = {
		ArrayExpression: ArrayExpression,
		ArrowFunctionExpression: ArrowFunctionExpression,
		AssignmentExpression: AssignmentExpression,
		BinaryExpression: BinaryExpression,
		BreakStatement: BreakStatement,
		CallExpression: CallExpression,
		ClassBody: ClassBody,
		ClassDeclaration: ClassDeclaration,
		ClassExpression: ClassExpression,
		ContinueStatement: ContinueStatement,
		DoWhileStatement: LoopStatement,
		ExportNamedDeclaration: ExportNamedDeclaration,
		ExportDefaultDeclaration: ExportDefaultDeclaration,
		ForStatement: ForStatement,
		ForInStatement: ForInStatement,
		ForOfStatement: ForOfStatement,
		FunctionDeclaration: FunctionDeclaration,
		FunctionExpression: FunctionExpression,
		Identifier: Identifier,
		ImportDeclaration: ImportDeclaration,
		ImportDefaultSpecifier: ImportDefaultSpecifier,
		ImportSpecifier: ImportSpecifier,
		Literal: Literal,
		MemberExpression: MemberExpression,
		Property: Property,
		ReturnStatement: ReturnStatement,
		Super: Super,
		TaggedTemplateExpression: TaggedTemplateExpression,
		TemplateElement: TemplateElement,
		TemplateLiteral: TemplateLiteral,
		ThisExpression: ThisExpression,
		UpdateExpression: UpdateExpression,
		VariableDeclaration: VariableDeclaration,
		VariableDeclarator: VariableDeclarator,
		WhileStatement: LoopStatement
	};

	function wrap ( raw, parent ) {
		if ( Array.isArray( raw ) ) {
			return raw.map( function ( value ) { return wrap( value, parent ); } );
		}

		if ( raw && typeof raw === 'object' ) {
			if ( raw.type === 'BlockStatement' ) return new BlockStatement( raw, parent );

			var Constructor = types[ raw.type ] || Node;
			return new Constructor( raw, parent );
		}

		return raw; // scalar value
	}

	var letConst = /^(?:let|const)$/;

	function Scope ( options ) {
		options = options || {};

		this.parent = options.parent;
		this.isBlockScope = !!options.block;

		var scope = this;
		while ( scope.isBlockScope ) scope = scope.parent;
		this.functionScope = scope;

		this.declarations = Object.create( null );
		this.references = Object.create( null );
		this.blockScopedDeclarations = this.isBlockScope ? null : Object.create( null );
		this.aliases = this.isBlockScope ? null : Object.create( null );
	}

	Scope.prototype = {
		addDeclaration: function addDeclaration ( node, kind ) {
			for ( var i = 0, list = extractNames( node ); i < list.length; i += 1 ) {
				var name = list[i];

				var existingDeclaration = this.declarations[ name ];
				if ( existingDeclaration && ( letConst.test( kind ) || letConst.test( existingDeclaration.kind ) ) ) {
					// TODO warn about double var declarations?
					throw new CompileError( node, ("" + name + " is already declared") );
				}

				var declaration = { name: name, node: node, kind: kind, instances: [] };
				this.declarations[ name ] = declaration;

				if ( this.isBlockScope ) {
					if ( !this.functionScope.blockScopedDeclarations[ name ] ) this.functionScope.blockScopedDeclarations[ name ] = [];
					this.functionScope.blockScopedDeclarations[ name ].push( declaration );
				}
			}
		},

		addReference: function addReference ( identifier ) {
			var declaration = this.declarations[ identifier.name ];
			if ( declaration ) {
				declaration.instances.push( identifier );
			} else {
				this.references[ identifier.name ] = true;
				if ( this.parent ) this.parent.addReference( identifier );
			}
		},

		contains: function contains ( name ) {
			return this.declarations[ name ] ||
			       ( this.parent ? this.parent.contains( name ) : false );
		},

		createIdentifier: function createIdentifier ( base ) {
			var this$1 = this;

			var name = base;
			var counter = 1;

			while ( this$1.declarations[ name ] || this$1.references[ name ] || this$1.aliases[ name ] || name in reserved ) {
				name = "" + base + "$" + (counter++);
			}

			this.aliases[ name ] = true;
			return name;
		},

		findDeclaration: function findDeclaration ( name ) {
			return this.declarations[ name ] ||
			       ( this.parent && this.parent.findDeclaration( name ) );
		}
	};

	var BlockStatement = (function (Node) {
		function BlockStatement () {
			Node.apply(this, arguments);
		}

		BlockStatement.prototype = Object.create( Node && Node.prototype );
		BlockStatement.prototype.constructor = BlockStatement;

		BlockStatement.prototype.createScope = function createScope () {
			var this$1 = this;

			this.parentIsFunction = /Function/.test( this.parent.type );
			this.isFunctionBlock = this.parentIsFunction || this.parent.type === 'Root';
			this.scope = new Scope({
				block: !this.isFunctionBlock,
				parent: this.parent.findScope( false )
			});

			if ( this.parentIsFunction ) {
				this.parent.params.forEach( function ( node ) {
					this$1.scope.addDeclaration( node, 'param' );
				});
			}
		};

		BlockStatement.prototype.initialise = function initialise ( transforms ) {
			this.thisAlias = null;
			this.argumentsAlias = null;
			this.defaultParameters = [];

			// normally the scope gets created here, during initialisation,
			// but in some cases (e.g. `for` statements), we need to create
			// the scope early, as it pertains to both the init block and
			// the body of the statement
			if ( !this.scope ) this.createScope();

			this.body.forEach( function ( node ) { return node.initialise( transforms ); } );
		};

		BlockStatement.prototype.findLexicalBoundary = function findLexicalBoundary () {
			if ( this.type === 'Program' ) return this;
			if ( /^Function/.test( this.parent.type ) ) return this;

			return this.parent.findLexicalBoundary();
		};

		BlockStatement.prototype.findScope = function findScope ( functionScope ) {
			if ( functionScope && !this.isFunctionBlock ) return this.parent.findScope( functionScope );
			return this.scope;
		};

		BlockStatement.prototype.getArgumentsAlias = function getArgumentsAlias () {
			if ( !this.argumentsAlias ) {
				this.argumentsAlias = this.scope.createIdentifier( 'arguments' );
			}

			return this.argumentsAlias;
		};

		BlockStatement.prototype.getThisAlias = function getThisAlias () {
			if ( !this.thisAlias ) {
				this.thisAlias = this.scope.createIdentifier( 'this' );
			}

			return this.thisAlias;
		};

		BlockStatement.prototype.transpile = function transpile ( code, transforms ) {
			var this$1 = this;

			var start = this.parent.type === 'Root' || this.synthetic ? this.start : this.start + 1;

			var indentation = this.synthetic ?
				this.getIndentation() + code.getIndentString() :
				( this.body.length ? this.body[0].getIndentation() : '' );

			var introStatementGenerators = [];

			if ( this.argumentsAlias ) {
				introStatementGenerators.push( function ( start, prefix, suffix ) {
					var assignment = "" + prefix + "var " + (this$1.argumentsAlias) + " = arguments;" + suffix;
					code.insertLeft( start, assignment );
				});
			}

			if ( this.thisAlias ) {
				introStatementGenerators.push( function ( start, prefix, suffix ) {
					var assignment = "" + prefix + "var " + (this$1.thisAlias) + " = this;" + suffix;
					code.insertLeft( start, assignment );
				});
			}

			if ( /Function/.test( this.parent.type ) ) {
				// default parameters
				if ( transforms.defaultParameter ) {
					this.transpileDefaultParameters( code, introStatementGenerators );
				}

				// object pattern
				if ( transforms.parameterDestructuring ) {
					this.transpileObjectPattern( code, introStatementGenerators );
					this.transpileArrayPattern( code, introStatementGenerators );
				}

				// rest parameter
				if ( transforms.spreadRest ) {
					this.transpileRestElement( code, introStatementGenerators, indentation );
				}
			}

			if ( transforms.letConst && this.isFunctionBlock ) {
				this.transpileBlockScopedIdentifiers( code );
			}

			Node.prototype.transpile.call( this, code, transforms );

			if ( this.synthetic ) {
				if ( this.parent.type === 'ArrowFunctionExpression' ) {
					var expr = this.body[0];

					if ( introStatementGenerators.length ) {
						code.insertLeft( this.start, ("{") ).insertRight( this.end, ("" + (this.parent.getIndentation()) + "}") );

						code.insertRight( expr.start, ("\n" + indentation + "return ") );
						code.insertLeft( expr.end, (";\n") );
					} else if ( transforms.arrow ) {
						code.insertRight( expr.start, ("{ return ") );
						code.insertLeft( expr.end, ("; }") );
					}
				}

				else if ( introStatementGenerators.length ) {
					code.insertLeft( this.start, ("{") ).insertRight( this.end, ("}") );
				}
			}

			var prefix = "\n" + indentation;
			var suffix = '';
			introStatementGenerators.forEach( function ( fn, i ) {
				if ( i === introStatementGenerators.length - 1 ) suffix = ( this$1.parent.type === 'Root' ? ("\n") : ("\n") );
				fn( start, prefix, suffix );
			});
		};

		BlockStatement.prototype.transpileBlockScopedIdentifiers = function transpileBlockScopedIdentifiers ( code ) {
			var this$1 = this;

			Object.keys( this.scope.blockScopedDeclarations ).forEach( function ( name ) {
				var declarations = this$1.scope.blockScopedDeclarations[ name ];

				for ( var i = 0; i < declarations.length; i += 1 ) {
					var declaration = declarations[i];
					var cont = false; // TODO implement proper continue...

					if ( declaration.kind === 'for.let' ) {
						// special case
						var forStatement = declaration.node.findNearest( 'ForStatement' );

						if ( forStatement.shouldRewriteAsFunction ) {
							var outerAlias = this$1.scope.createIdentifier( name );
							var innerAlias = forStatement.reassigned[ name ] ?
								this$1.scope.createIdentifier( name ) :
								name;

							declaration.name = outerAlias;

							forStatement.aliases[ name ] = {
								outer: outerAlias,
								inner: innerAlias
							};

							for ( var i$1 = 0, list = declaration.instances; i$1 < list.length; i$1 += 1 ) {
								var identifier = list[i$1];

								var alias = forStatement.body.contains( identifier ) ?
									innerAlias :
									outerAlias;

								if ( name !== alias ) {
									code.overwrite( identifier.start, identifier.end, alias, true );
								}
							}

							cont = true;
						}
					}

					if ( !cont ) {
						var alias$1 = this$1.scope.createIdentifier( name );

						if ( name !== alias$1 ) {
							declaration.name = alias$1;

							for ( var i$2 = 0, list$1 = declaration.instances; i$2 < list$1.length; i$2 += 1 ) {
								var identifier$1 = list$1[i$2];

								identifier$1.rewritten = true;
								code.overwrite( identifier$1.start, identifier$1.end, alias$1, true );
							}
						}
					}
				}
			});
		};

		BlockStatement.prototype.transpileDefaultParameters = function transpileDefaultParameters ( code, introStatementGenerators ) {
			this.parent.params.filter( function ( param ) { return param.type === 'AssignmentPattern'; } ).forEach( function ( param ) {
				introStatementGenerators.push( function ( start, prefix, suffix ) {
					var lhs = "" + prefix + "if ( " + (param.left.name) + " === void 0 ) " + (param.left.name);
					code
						.insertRight( param.left.end, ("" + lhs) )
						.move( param.left.end, param.right.end, start )
						.insertLeft( param.right.end, (";" + suffix) );
				});
			});
		};

		BlockStatement.prototype.transpileObjectPattern = function transpileObjectPattern ( code, introStatementGenerators ) {
			var this$1 = this;

			this.parent.params.filter( function ( param ) { return param.type === 'ObjectPattern'; } ).forEach( function ( param ) {
				var ref = this$1.scope.createIdentifier( 'ref' );
				code.insertRight( param.start, ref );

				var c = param.start;

				param.properties.forEach( function ( prop ) {
					code.remove( c, prop.value.start );
					var key = prop.key.name;

					if ( prop.value.type === 'Identifier' ) {
						code.remove( prop.value.start, prop.value.end );

						var value = prop.value.name;
						var declaration = this$1.scope.findDeclaration( value );

						if ( declaration.instances.length === 1 ) {
							var instance = declaration.instances[0];
							code.overwrite( instance.start, instance.end, ("" + ref + "." + key) );
						} else {
							introStatementGenerators.push( function ( start, prefix, suffix ) {
								code.insertLeft( start, ("" + prefix + "var " + value + " = " + ref + "." + key + ";" + suffix) );
							});
						}
					}

					else if ( prop.value.type === 'AssignmentPattern' ) {
						introStatementGenerators.push( function ( start, prefix, suffix ) {
							code.remove( prop.value.start, prop.value.right.start );

							var value = prop.value.left.name;

							code
								.insertRight( prop.value.right.start, ("" + prefix + "var " + ref + "_" + key + " = " + ref + "." + key + ", " + value + " = " + ref + "_" + key + " === void 0 ? ") )
								.insertLeft( prop.value.right.end, (" : " + ref + "_" + key + ";" + suffix) )
								.move( prop.value.right.start, prop.value.right.end, start );
						});
					}

					else {
						throw new CompileError( prop, ("Compound destructuring is not supported") );
					}

					c = prop.value.end;
				});

				code.remove( c, param.end );
			});
		};

		BlockStatement.prototype.transpileArrayPattern = function transpileArrayPattern ( code, introStatementGenerators ) {
			// array pattern. TODO dry this out
			var this$1 = this;

			this.parent.params.filter( function ( param ) { return param.type === 'ArrayPattern'; } ).forEach( function ( param ) {
				var ref = this$1.scope.createIdentifier( 'ref' );
				code.insertRight( param.start, ref );

				var c = param.start;

				param.elements.forEach( function ( element, i ) {
					code.remove( c, element.start );

					if ( element.type === 'Identifier' ) {
						code.remove( element.start, element.end );

						introStatementGenerators.push( function ( start, prefix, suffix ) {
							code.insertLeft( start, ("" + prefix + "var " + (element.name) + " = " + ref + "[" + i + "];" + suffix) );
						});
					} else if ( element.type === 'AssignmentPattern' ) {
						introStatementGenerators.push( function ( start, prefix, suffix ) {
							code.remove( element.start, element.right.start );

							var name = element.left.name;
							code
								.insertRight( element.right.start, ("" + prefix + "var " + ref + "_" + i + " = ref[" + i + "], " + name + " = ref_" + i + " === void 0 ? ") )
								.insertLeft( element.right.end, (" : ref_" + i + ";" + suffix) )
								.move( element.right.start, element.right.end, start );
						});
					}

					else {
						throw new CompileError( element, ("Compound destructuring is not supported") );
					}

					c = element.end;
				});

				code.remove( c, param.end );
			});
		};

		BlockStatement.prototype.transpileRestElement = function transpileRestElement ( code, introStatementGenerators, indentation ) {
			var this$1 = this;

			var params = this.parent.params;

			var lastParam = params[ params.length - 1 ];
			if ( lastParam && lastParam.type === 'RestElement' ) {
				introStatementGenerators.push( function ( start, prefix, suffix ) {
					var penultimateParam = params[ params.length - 2 ];

					if ( penultimateParam ) {
						code.remove( penultimateParam ? penultimateParam.end : lastParam.start, lastParam.end );
					} else {
						var start$1 = lastParam.start, end = lastParam.end; // TODO https://gitlab.com/Rich-Harris/buble/issues/8

						while ( /\s/.test( code.original[ start$1 - 1 ] ) ) start$1 -= 1;
						while ( /\s/.test( code.original[ end ] ) ) end += 1;

						code.remove( start$1, end );
					}

					var name = lastParam.argument.name;
					var len = this$1.scope.createIdentifier( 'len' );
					var count = params.length - 1;

					if ( count ) {
						code.insertLeft( start, ("" + prefix + "var " + name + " = [], " + len + " = arguments.length - " + count + ";\n" + indentation + "while ( " + len + "-- > 0 ) " + name + "[ " + len + " ] = arguments[ " + len + " + " + count + " ];" + suffix) );
					} else {
						code.insertLeft( start, ("" + prefix + "var " + name + " = [], " + len + " = arguments.length;\n" + indentation + "while ( " + len + "-- ) " + name + "[ " + len + " ] = arguments[ " + len + " ];" + suffix) );
					}
				});
			}
		};

		return BlockStatement;
	}(Node));

	function Program ( source, ast, transforms ) {
		var this$1 = this;

		this.type = 'Root';

		this.source = source;
		this.magicString = new MagicString( source );

		this.ast = ast;
		this.depth = 0;

		this.body = new BlockStatement( ast, this );

		this.templateElements = [];
		this.body.initialise( transforms );

		this.indentExclusions = {};
		for ( var i$1 = 0, list = this.templateElements; i$1 < list.length; i$1 += 1 ) {
			var node = list[i$1];

			for ( var i = node.start; i < node.end; i += 1 ) {
				this$1.indentExclusions[ node.start + i ] = true;
			}
		}

		this.body.transpile( this.magicString, transforms );
	}

	Program.prototype = {
		export: function export$1 ( options ) {
			if ( options === void 0 ) options = {};

			return {
				code: this.magicString.toString(),
				map: this.magicString.generateMap({
					file: options.file,
					source: options.source,
					includeContent: options.includeContent !== false
				})
			};
		},

		findNearest: function findNearest () {
			return null;
		},

		findScope: function findScope () {
			return null;
		}
	};

	var matrix = {
		chrome: {
			    48: 1333689725,
			    49: 1342078975,
			    50: 1610514431
		},
		firefox: {
			    43: 1207307741,
			    44: 1207307741,
			    45: 1207307741
		},
		safari: {
			     8: 1073741824,
			     9: 1328940894
		},
		ie: {
			     8: 0,
			     9: 1073741824,
			    10: 1073741824,
			    11: 1073770592
		},
		edge: {
			    12: 1591620701,
			    13: 1608400479
		},
		node: {
			'0.10': 1075052608,
			'0.12': 1091830852,
			     4: 1327398527,
			     5: 1327398527,
			     6: 1610514431
		}
	};

	var features = [
		'arrow',
		'classes',
		'collections',
		'computedProperty',
		'conciseMethodProperty',
		'constLoop',
		'constRedef',
		'defaultParameter',
		'destructuring',
		'extendNatives',
		'forOf',
		'generator',
		'letConst',
		'letLoop',
		'letLoopScope',
		'moduleExport',
		'moduleImport',
		'numericLiteral',
		'objectProto',
		'objectSuper',
		'oldOctalLiteral',
		'parameterDestructuring',
		'spreadRest',
		'stickyRegExp',
		'symbol',
		'templateString',
		'unicodeEscape',
		'unicodeIdentifier',
		'unicodeRegExp',

		// ES2016
		'exponentiation',

		// additional transforms, not from
		// https://featuretests.io
		'reservedProperties'
	];

	var version = "0.8.2";

	var dangerousTransforms = [
		'dangerousTaggedTemplateString',
		'dangerousForOf'
	];

	function target ( target ) {
		var targets = Object.keys( target );
		var bitmask = targets.length ?
			2147483647 :
			1073741824;

		Object.keys( target ).forEach( function ( environment ) {
			var versions = matrix[ environment ];
			if ( !versions ) throw new Error( ("Unknown environment '" + environment + "'. Please raise an issue at https://gitlab.com/Rich-Harris/buble/issues") );

			var targetVersion = target[ environment ];
			if ( !( targetVersion in versions ) ) throw new Error( ("Support data exists for the following versions of " + environment + ": " + (Object.keys( versions ).join( ', ')) + ". Please raise an issue at https://gitlab.com/Rich-Harris/buble/issues") );
			var support = versions[ targetVersion ];

			bitmask &= support;
		});

		var transforms = Object.create( null );
		features.forEach( function ( name, i ) {
			transforms[ name ] = !( bitmask & 1 << i );
		});

		dangerousTransforms.forEach( function ( name ) {
			transforms[ name ] = false;
		});

		return transforms;
	}

	function transform ( source, options ) {
		if ( options === void 0 ) options = {};

		var ast;

		try {
			ast = acorn.parse( source, {
				ecmaVersion: 7,
				preserveParens: true,
				sourceType: 'module'
			});
		} catch ( err ) {
			err.snippet = getSnippet( source, err.loc );
			throw err;
		}

		var transforms = target( options.target || {} );
		Object.keys( options.transforms || {} ).forEach( function ( name ) {
			if ( name === 'modules' ) {
				if ( !( 'moduleImport' in options.transforms ) ) transforms.moduleImport = options.transforms.modules;
				if ( !( 'moduleExport' in options.transforms ) ) transforms.moduleExport = options.transforms.modules;
				return;
			}

			if ( !( name in transforms ) ) throw new Error( ("Unknown transform '" + name + "'") );
			transforms[ name ] = options.transforms[ name ];
		});

		return new Program( source, ast, transforms ).export( options );
	}

	exports.target = target;
	exports.transform = transform;
	exports.VERSION = version;

}));
//# sourceMappingURL=buble.umd.js.map